"""Provides classes for general GetStimStatus"""

from typing import NamedTuple
from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.packet import Packet, PacketAck
from science_mode_4.protocol.types import StimStatus

class GetStimStatusResult(NamedTuple):
    """Helper class for stim result"""
    stim_status: StimStatus
    high_voltage_on: bool


class PacketGeneralGetStimStatus(Packet):
    """Packet for general GetStimStatus"""


    def __init__(self):
        super().__init__()
        self._command = Commands.GET_STIM_STATUS


class PacketGeneralGetStimStatusAck(PacketAck):
    """Packet for general GetStimStatus akcnowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.GET_STIM_STATUS_ACK
        self._successful: bool = False
        self._stim_status = StimStatus.NO_LEVEL_INITIALIZED
        self._high_voltage_on: bool = False

        if not data is None:
            self._successful = data[0] == 0
            self._stim_status = StimStatus(data[1])
            self._high_voltage_on = data[2] == 6


    @property
    def successful(self) -> bool:
        """Getter for Successful"""
        return self._successful


    @property
    def stim_status(self) -> StimStatus:
        """Getter for StimStatus"""
        return self._stim_status


    @property
    def high_voltage_on(self) -> bool:
        """Getter for HighVoltageOn"""
        return self._high_voltage_on



"""Provides classes for mid level Update"""

from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.packet import Packet, PacketAck
from science_mode_4.protocol.types import ResultAndError
from science_mode_4.utils.byte_builder import ByteBuilder
from .mid_level_types import MidLevelChannelConfiguration


class PacketMidLevelUpdate(Packet):
    """Packet for mid level Update"""


    def __init__(self):
        super().__init__()
        self._command = Commands.MID_LEVEL_UPDATE
        self._channel_configuration: list[MidLevelChannelConfiguration] = [None] * 8


    @property
    def channel_configuration(self) -> list[MidLevelChannelConfiguration]:
        """Getter for channel configuration"""
        return self._channel_configuration


    @channel_configuration.setter
    def channel_configuration(self, channel_configuration: list[MidLevelChannelConfiguration]):
        """Setter for channel configuration"""
        self._channel_configuration = channel_configuration


    def get_data(self) -> bytes:
        bb = ByteBuilder()
        for x in range(8):
            c: MidLevelChannelConfiguration | None = self._channel_configuration[x] if x < len(self._channel_configuration) else None
            bb.set_bit_to_position(1 if c and c.is_active else 0, x, 1)

        for x in range(8):
            c: MidLevelChannelConfiguration | None = self._channel_configuration[x] if x < len(self._channel_configuration) else None
            if c:
                bb.append_bytes(c.get_data())

        return bb.get_bytes()


class PacketMidLevelUpdateAck(PacketAck):
    """Packet for mid level Update acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.MID_LEVEL_UPDATE_ACK
        self._result_error = ResultAndError.NO_ERROR

        if not data is None:
            self._result_error = ResultAndError(data[0])


    @property
    def result_error(self) -> ResultAndError:
        """Getter for ResultError"""
        return self._result_error



"""Provides a base class for a connection"""

from abc import ABC, abstractmethod


class Connection(ABC):
    """Abstract base class for connection"""


    @abstractmethod
    def open(self):
        """Open connection"""


    @abstractmethod
    def close(self):
        """Close connection"""


    @abstractmethod
    def is_open(self) -> bool:
        """Checks if connection is open"""


    @abstractmethod
    def write(self, data: bytes):
        """Write data to connection"""


    @abstractmethod
    def read(self) -> bytes:
        """Read all data from connection"""


    @abstractmethod
    def clear_buffer(self):
        """Clear buffer from connection"""



"""Provides packet classes for dyscom get with type file system status"""

from typing import NamedTuple

from .dyscom_types import DyscomFileByNameMode, DyscomGetType
from .dyscom_helper import DyscomHelper
from .dyscom_get import PacketDyscomGet, PacketDyscomGetAck


class DyscomGetFileByNameResult(NamedTuple):
    """Helper class for dyscom get with type file by name"""
    filename: str
    block_offset: int
    filesize: int
    number_of_blocks: int
    mode: DyscomFileByNameMode


class PacketDyscomGetFileByName(PacketDyscomGet):
    """Packet for dyscom get with type file by name"""


    def __init__(self):
        super().__init__()
        self._type = DyscomGetType.FILE_BY_NAME
        self._kind = int(self._type)


class PacketDyscomGetAckFileByName(PacketDyscomGetAck):
    """Packet for dyscom get for type file by name acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._kind = int(DyscomGetType.FILE_BY_NAME)
        self._filename = ""
        self._block_offset = 0
        self._filesize = 0
        self._number_of_blocks = 0
        self._mode: DyscomFileByNameMode.UNDEFINED

        if not data is None:
            self._filename = DyscomHelper.bytes_to_str(data[2:130], 128)
            self._block_offset = int.from_bytes(data[130:134], "little")
            self._filesize = int.from_bytes(data[134:142], "little")
            self._number_of_blocks = int.from_bytes(data[142:146], "little")
            self._mode = DyscomFileByNameMode(data[146])



    @property
    def filename(self) -> str:
        """Getter for filename"""
        return self._filename


    @property
    def block_offset(self) -> int:
        """Getter for block offset"""
        return self._block_offset


    @property
    def filesize(self) -> int:
        """Getter for filesize"""
        return self._filesize


    @property
    def number_of_blocks(self) -> int:
        """Getter for number of blocks"""
        return self._number_of_blocks


    @property
    def mode(self) -> DyscomFileByNameMode:
        """Getter for mode"""
        return self._mode



"""Provides device class representing a P24"""

from .device import Device, DeviceCapability
from .utils.connection import Connection


class DeviceP24(Device):
    """Device class for a P24 device"""

    def __init__(self, conn: Connection):
        super().__init__(conn, [DeviceCapability.GENERAL,
                                DeviceCapability.LOW_LEVEL,
                                DeviceCapability.MID_LEVEL])



"""Example how to use dyscom and write values to a .csv file.
Because writing takes time, we handle it in a
background thread and synchronize measurement data vie Queue from main thread"""

import asyncio
from timeit import default_timer as timer

from science_mode_4 import DeviceI24
from science_mode_4 import Commands
from science_mode_4 import SerialPortConnection
from science_mode_4.dyscom.ads129x.ads129x_config_register_1 import Ads129xOutputDataRate, Ads129xPowerMode
from science_mode_4.dyscom.dyscom_get_operation_mode import PacketDyscomGetAckOperationMode
from science_mode_4.dyscom.dyscom_send_live_data import PacketDyscomSendLiveData
from science_mode_4.dyscom.dyscom_types import DyscomFilterType, DyscomGetType, DyscomInitParams, DyscomPowerModulePowerType,\
    DyscomPowerModuleType, DyscomSignalType
from science_mode_4.protocol.types import ResultAndError
from science_mode_4.utils.logger import logger
from examples.utils.example_utils import ExampleUtils
from examples.utils.csv_utils import CsvHelper


def main():
    """Main function"""

    csv_helper = CsvHelper("values.csv", ["package_nr", "bi", "emg_1", "emg_2", "breathing", "temperature", "time_delta"])
    csv_helper.start()

    async def device_communication() -> int:
        """Communication with science mode device"""

        # disable logger to increase performance
        logger().disabled = True

        # get comport from command line argument
        com_port = ExampleUtils.get_comport_from_commandline_argument()
        # create serial port connection
        connection = SerialPortConnection(com_port)
        # open connection, now we can read and write data
        connection.open()

        # create science mode device
        device = DeviceI24(connection)
        # call initialize to get basic information (serial, versions) and stop any active stimulation/measurement
        # to have a defined state
        await device.initialize()

        # get dyscom layer to call dyscom level commands
        dyscom = device.get_layer_dyscom()

        # call enable measurement power module for measurement
        await dyscom.power_module(DyscomPowerModuleType.MEASUREMENT, DyscomPowerModulePowerType.SWITCH_ON)
        # call init with 4k sample rate and enable signal types
        init_params = DyscomInitParams()
        init_params.filter = DyscomFilterType.PREDEFINED_FILTER_3
        init_params.signal_type = [DyscomSignalType.BI, DyscomSignalType.EMG_1,\
                                DyscomSignalType.EMG_2, DyscomSignalType.BREATHING, DyscomSignalType.TEMPERATURE]
        init_params.register_map_ads129x.config_register_1.output_data_rate = Ads129xOutputDataRate.HR_MODE_4_KSPS__LP_MODE_2_KSPS
        init_params.register_map_ads129x.config_register_1.power_mode = Ads129xPowerMode.HIGH_RESOLUTION
        await dyscom.init(init_params)

        # start dyscom measurement
        await dyscom.start()

        start_time = timer()
        total_count = 0

        # loop for some time
        for x in range(1000):
            # check operation mode from time to time, this function is not waiting for response
            # so we have to handle it by ourself later
            if x % 100 == 0:
                dyscom.send_get_operation_mode()

            live_data_counter = 0
            while True:
                # process all available packages
                ack = dyscom.packet_buffer.get_packet_from_buffer(live_data_counter == 0)
                if ack:
                    # because there are multiple get commands, we need to additionally check kind,
                    # which is always associated DyscomGetType
                    if ack.command == Commands.DL_GET_ACK and ack.kind == DyscomGetType.OPERATION_MODE:
                        om_ack: PacketDyscomGetAckOperationMode = ack
                        print(f"Operation mode {om_ack.operation_mode.name}")
                        # check if measurement is still active
                        if om_ack.result_error != ResultAndError.NO_ERROR:
                            break
                    elif ack.command == Commands.DL_SEND_LIVE_DATA:
                        live_data_counter += 1
                        total_count += 1

                        sld: PacketDyscomSendLiveData = ack
                        if sld.status_error:
                            print(f"SendLiveData status error {sld.samples}")
                            break

                        csv_helper.append_values(ack.number, [sld.samples[0].value, sld.samples[1].value,\
                                                              sld.samples[2].value, sld.samples[3].value,\
                                                              sld.samples[4].value], sld.time_offset)

                else:
                    # print(f"Live data acknowledges per iteration {live_data_counter}")
                    break

            # await asyncio.sleep(0.01)

        # print stats
        end_time = timer()
        print(f"Samples: {total_count}, duration: {end_time - start_time}, sample rate: {total_count / (end_time - start_time)}")

        # stop measurement
        await dyscom.stop()
        # turn power module off
        await dyscom.power_module(DyscomPowerModuleType.MEASUREMENT, DyscomPowerModulePowerType.SWITCH_OFF)

        # close serial port connection
        connection.close()

        # stop writing file
        csv_helper.stop()

        return 0


    # start device communication
    asyncio.run(device_communication())


if __name__ == "__main__":
    main()



"""Provides dyscom types"""

from dataclasses import dataclass, field
from enum import IntEnum
import datetime

from science_mode_4.utils.byte_builder import ByteBuilder
from .ads129x.ads129x import Ads129x
from .dyscom_helper import DyscomHelper

class DyscomFrequencyOut(IntEnum):
    """Represent dyscom type for frequency out"""
    UNUSED = 0
    SAMPLES_PER_SECOND_32K = 1
    SAMPLES_PER_SECOND_16K = 2
    SAMPLES_PER_SECOND_8K = 3
    SAMPLES_PER_SECOND_4K = 4
    SAMPLES_PER_SECOND_2K = 5
    SAMPLES_PER_SECOND_1K = 6
    SAMPLES_PER_SECOND_500 = 7
    SAMPLES_PER_SECOND_250  = 8


class DyscomInitState(IntEnum):
    """Represent dyscom type for init state"""
    UNUSED = 0
    SUCCESS = 1
    ERROR_STORAGE_INIT = 2
    ERROR_STORAGE_WRITE = 3
    ERROR_STORAGE_FULL = 4
    UNUSED2 = 5
    ERROR_ADS129X_REGISTER  = 6


class DyscomSignalType(IntEnum):
    """Represent dyscom type for signal type"""
    UNUSED = 0
    UNKNOWN = 1
    BI = 2
    EMG_1 = 3
    OP_VOLTAGE = 4
    TEST_SIGNAL = 5
    GROUND = 6
    TEMPERATURE = 7
    INTERNAL_SC = 8
    EMG_2 = 9
    TIME = 10
    PUSHBUTTON = 11
    BREATHING = 12


class DyscomFilterType(IntEnum):
    """Represent dyscom type for filter type"""
    FILTER_OFF = 0
    PREDEFINED_FILTER_1 = 1
    PREDEFINED_FILTER_2 = 2
    PREDEFINED_FILTER_3 = 3


class DyscomInitFlag(IntEnum):
    """Represent dyscom type for init flag"""
    ENABLE_LIVE_DATA_MODE = 0
    ENABLE_SD_STORAGE_MODE = 1
    ENABLE_TIMED_START = 2
    SET_SYSTEM_TIME_WITH_SENDED_SYSTEM_TIME_STAMP = 3
    MUTE = 4


class DyscomGetType(IntEnum):
    """Represent dyscom type for get"""
    BATTERY = 0
    FILESYSTEM_STATUS = 1
    LIST_OF_MEASUREMENT_META_INFO = 2
    OPERATION_MODE = 3
    FILE_BY_NAME = 4
    DEVICE_ID = 5
    FIRMWARE_VERSION = 6
    FILE_INFO = 7


class DyscomGetOperationModeType(IntEnum):
    """Represents dyscom get operation mode type"""
    UNDEFINED = 0
    IDLE = 1
    LIVE_MEASURING_PRE = 2
    LIVE_MEASURING = 3
    RECORD_PRE = 4
    RECORD = 5
    DATATRANSFER_PRE = 6
    DATATRANSFER = 7


class DyscomPowerModuleType(IntEnum):
    """Represents dyscom power module type"""
    BLUETOOTH = 1
    MEMORY_CARD = 2
    MEASUREMENT = 3
    RESEARCH = 4


class DyscomPowerModulePowerType(IntEnum):
    """Represents dyscom power module power type"""
    SWITCH_OFF = 0
    SWITCH_ON = 1


class DyscomPowerLiveDataStatusFlag(IntEnum):
    """Represents dyscom live data status type"""
    POSITIVE_ELECTRODE_ADHESIVE = 1
    NEGATIVE_ELECTRODE_ADHESIVE = 2
    BOTH_ELECTRODES_ADHESIVE = 3


class DyscomFileByNameMode(IntEnum):
    """Represents dyscom file by name mode type"""
    UNDEFINED = 0
    MULTIBLOCK = 1
    SINGLEBLOCK = 2


class DyscomEnergyFlag(IntEnum):
    """Represents dyscom energy flag type"""
    UNDEFINED = 0
    CABLE_CONNECTED = 1
    DEVICE_IS_LOADING = 2


class DyscomSysType(IntEnum):
    """Represents dyscom sys type"""
    UNDEFINED = 0
    DELETE_FILE = 1
    DEVICE_SLEEP = 2
    DEVICE_STORAGE = 3


class DyscomSysState(IntEnum):
    """Represents dyscom sys state"""
    UNDEFINED = 0
    SUCCESSFUL = 1


@dataclass
class DyscomElectrodeSample:
    """Represent an electrode sample send by live data packet"""
    value: float = 0.0
    signal_type: DyscomSignalType = DyscomSignalType.UNUSED
    status: set[DyscomPowerLiveDataStatusFlag] = field(default_factory=lambda: set()) # pylint:disable=unnecessary-lambda


@dataclass
class DyscomInitParams():
    """Dyscom init packet parameters"""

    register_map_ads129x = Ads129x()
    start_time = datetime.datetime.now()
    system_time = datetime.datetime.now()
    proband_name = ""
    investigator_name = ""
    proband_number = ""
    duration = datetime.timedelta()
    signal_type: list[DyscomSignalType] = field(default_factory=lambda: [DyscomSignalType.BI,  DyscomSignalType.EMG_1])
    sync_signal: bool = False
    filter = DyscomFilterType.FILTER_OFF
    flags: set[DyscomInitFlag] = field(default_factory=lambda: {DyscomInitFlag.ENABLE_LIVE_DATA_MODE})


    def get_data(self) -> bytes:
        """Convert information to bytes"""

        if len(self.proband_name) > 128:
            raise ValueError(f"Proband name must be shorter than 129 {self.proband_name}")
        if len(self.investigator_name) > 128:
            raise ValueError(f"Investigator name must be shorter than 129 {self.investigator_name}")
        if len(self.proband_number) > 36:
            raise ValueError(f"Proband number must be shorter than 37 {self.proband_number}")


        bb = ByteBuilder()
        bb.append_bytes(self.register_map_ads129x.get_data())
        bb.append_bytes(DyscomHelper.datetime_to_bytes(self.start_time))
        bb.append_bytes(DyscomHelper.datetime_to_bytes(self.system_time))
        bb.append_bytes(DyscomHelper.str_to_bytes(self.proband_name, 129))
        bb.append_bytes(DyscomHelper.str_to_bytes(self.investigator_name, 129))
        bb.append_bytes(DyscomHelper.str_to_bytes(self.proband_number, 37))
        bb.append_value(len(self.signal_type), 2, True)
        bb.append_value(self.duration.seconds, 4, True)
        for x in range(8):
            bb.append_byte(self.signal_type[x] if x < len(self.signal_type) else DyscomSignalType.UNUSED)
        bb.append_byte(0)
        bb.append_byte(1 if self.sync_signal else 0)
        bb.append_byte(self.filter)
        flags = 0
        for x in self.flags:
            flags |= 1 << x
        bb.append_byte(flags)

        return bb.get_bytes()


    def set_data(self, data: bytes):
        """Convert bytes to information"""



"""Provides packet classes for dyscom send file"""

import struct

from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.packet import Packet, PacketAck
from science_mode_4.utils.byte_builder import ByteBuilder


class PacketDyscomSendFile(PacketAck):
    """Packet for dyscom send file (this is technically not an acknowledge, but it is handled as such,
    because it is send automatically from device)"""


    _unpack_func = struct.Struct(">IH").unpack


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.DL_SEND_FILE
        self._block_number = 0
        self._block_size = 0
        self._data: bytes = bytes()

        if not data is None:
            self._block_number, self._block_size = PacketDyscomSendFile._unpack_func(">IH", data)
            self._data = data[6:self._block_size]


    @property
    def block_number(self) -> int:
        """Getter for block number"""
        return self._block_number


    @property
    def block_size(self) -> int:
        """Getter for block size"""
        return self._block_size


    @property
    def data(self) -> bytes:
        """Getter for data"""
        return self._data


class PacketDyscomSendFileAck(Packet):
    """Packet for dyscom send file acknowledge (this is technically not apacket, but it is handled as such,
    because it is send from PC to device)"""


    def __init__(self, block_number: int = 0):
        super().__init__()
        self._command = Commands.DL_SEND_FILE_ACK
        self._block_number = block_number


    def get_data(self) -> bytes:
        bb = ByteBuilder()
        bb.append_bytes(self._block_number)
        return bb.get_bytes()



"""Provides low level layer"""

from science_mode_4.layer import Layer
from science_mode_4.utils.logger import logger
from .dyscom_types import DyscomGetOperationModeType, DyscomPowerModuleType, DyscomPowerModulePowerType, DyscomSysType
from .dyscom_init import DyscomInitResult, PacketDyscomInit, PacketDyscomInitAck, DyscomInitParams
from .dyscom_get_file_system_status import PacketDyscomGetFileSystemStatus, PacketDyscomGetAckFileSystemStatus,\
    DyscomGetFileSystemStatusResult
from .dyscom_get_file_by_name import PacketDyscomGetFileByName, PacketDyscomGetAckFileByName, DyscomGetFileByNameResult
from .dyscom_get_firmware_version import PacketDyscomGetFirmwareVersion, PacketDyscomGetAckFirmwareVersion
from .dyscom_get_operation_mode import PacketDyscomGetOperationMode, PacketDyscomGetAckOperationMode
from .dyscom_start import PacketDyscomStart, PacketDyscomStartAck
from .dyscom_stop import PacketDyscomStop, PacketDyscomStopAck
from .dyscom_power_module import PacketDyscomPowerModule, PacketDyscomPowerModuleAck, DyscomPowerModuleResult
from .dyscom_get_list_of_measurement_meta_info import PacketDyscomGetAckListOfMeasurementMetaInfo, PacketDyscomGetListOfMeasurementMetaInfo
from .dyscom_get_device_id import PacketDyscomGetAckDeviceId, PacketDyscomGetDeviceId
from .dyscom_get_file_info import DyscomGetFileInfoResult, PacketDyscomGetAckFileInfo, PacketDyscomGetFileInfo
from .dyscom_get_battery_status import DyscomGetBatteryResult, PacketDyscomGetAckBatteryStatus, PacketDyscomGetBatteryStatus
from .dyscom_sys import DyscomSysResult, PacketDyscomSys, PacketDyscomSysAck
from .dyscom_send_file import PacketDyscomSendFileAck


class LayerDyscom(Layer):
    """
    Class for dyscom layer
    """


    async def init(self, params = DyscomInitParams()) -> DyscomInitResult:
        """Send dyscom init command and waits for response"""
        p = PacketDyscomInit(params)
        ack: PacketDyscomInitAck = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomInit")
        logger().info("Dyscom init, state: %s, frequency: %s", ack.init_state.name, ack.frequency_out.name)
        return DyscomInitResult(ack.register_map_ads129x, ack.init_state, ack.frequency_out)


    async def get_file_system_status(self) -> DyscomGetFileSystemStatusResult:
        """Sends get dyscom get type file system status and waits for response, returns file system ready, used size and free size"""
        p = PacketDyscomGetFileSystemStatus()
        ack: PacketDyscomGetAckFileSystemStatus = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomGetFileSystemStatus")
        logger().info("Dyscom get file system status, ready: %s, used size: %d, free size: %s",\
                      ack.file_system_ready, ack.used_size, ack.free_size)
        return DyscomGetFileSystemStatusResult(ack.file_system_ready, ack.used_size, ack.free_size)


    async def get_list_of_measurement_meta_info(self) -> int:
        """Sends get dyscom get type list of measurement meta info and waits for response, returns number of measurements"""
        p = PacketDyscomGetListOfMeasurementMetaInfo()
        ack: PacketDyscomGetAckListOfMeasurementMetaInfo = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomGetListOfMeasurementMetaInfo")
        logger().info("Dyscom get list of mmi, number of measurements: %d", ack.number_of_measurements)
        return ack.number_of_measurements


    async def get_file_by_name(self) -> DyscomGetFileByNameResult:
        """Sends get dyscom get type file by name and waits for response, returns filename, block offset,
        filesize, number of blocks and mode"""
        p = PacketDyscomGetFileByName()
        ack: PacketDyscomGetAckFileByName = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomGetFileByName")
        logger().info("Dyscom get file by name, filename: %s, block offset: %d, filesize: %d, number of blocks: %d, mode: %s",\
                      ack.filename, ack.block_offset, ack.filesize, ack.number_of_blocks, ack.mode.name)
        return DyscomGetFileByNameResult(ack.filename, ack.block_offset, ack.filesize, ack.number_of_blocks, ack.mode)


    async def get_device_id(self) -> str:
        """Sends get dyscom get type device id and waits for response, returns device id"""
        p = PacketDyscomGetDeviceId()
        ack: PacketDyscomGetAckDeviceId = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomGetDeviceId")
        logger().info("Dyscom get device id: %s", ack.device_id)
        return ack.device_id


    async def get_firmware_version(self) -> str:
        """Sends get dyscom get type firmware version and waits for response, returns firmware version"""
        p = PacketDyscomGetFirmwareVersion()
        ack: PacketDyscomGetAckFirmwareVersion = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomGetFirmwareVersion")
        logger().info("Dyscom get firmware version: %s", ack.firmware_version)
        return ack.firmware_version


    async def get_file_info(self) -> DyscomGetFileInfoResult:
        """Sends get dyscom get type file by name and waits for response, returns filename, block offset, filesize and number of blocks"""
        p = PacketDyscomGetFileInfo()
        ack: PacketDyscomGetAckFileInfo = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomGetFileInfo")
        logger().info("Dyscom get file info, filename: %s, filesize: %d, checksum: %d",\
                      ack.filename, ack.filesize, ack.checksum)
        return DyscomGetFileInfoResult(ack.filename, ack.filesize, ack.checksum)


    async def get_battery(self) -> DyscomGetBatteryResult:
        """Sends get dyscom get type batter and waits for response, returns voltage, current, percentage, temperature and energy state"""
        p = PacketDyscomGetBatteryStatus()
        ack: PacketDyscomGetAckBatteryStatus = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomGetBatteryStatus")
        logger().info("Dyscom get battery, voltage: %d, current: %d, percentage: %d, temperature: %d, energy state: %s",\
                      ack.voltage, ack.current, ack.percentage, ack.temperature, ack.energy_state)
        return DyscomGetBatteryResult(ack.voltage, ack.current, ack.percentage, ack.temperature, ack.energy_state)


    async def get_operation_mode(self) -> DyscomGetOperationModeType:
        """Sends get dyscom get type operation mode and waits for response, returns operation mode"""
        p = PacketDyscomGetOperationMode()
        ack: PacketDyscomGetAckOperationMode = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomGetFirmwareVersion")
        logger().info("Dyscom get operation mode: %s", ack.operation_mode.name)
        return ack.operation_mode


    def send_get_operation_mode(self):
        """Sends get dyscom get type operation mode and returns immediately without waiting for response"""
        p = PacketDyscomGetOperationMode()
        logger().info("Dyscom send get operation mode")
        self._send_packet(p)


    async def power_module(self, module: DyscomPowerModuleType, power: DyscomPowerModulePowerType) -> DyscomPowerModuleResult:
        """Sends get dyscom power module and waits for response, returns module and power"""
        p = PacketDyscomPowerModule(module, power)
        ack: PacketDyscomPowerModuleAck = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomStart")
        logger().info("Dyscom power module, module: %s, power: %s", ack.module.name, ack.power.name)
        return DyscomPowerModuleResult(ack.module, ack.power)


    async def sys(self, sys_type: DyscomSysType, filename: str = "") -> DyscomSysResult:
        """Sends get dyscom sys and waits for response, returns type, state and filename"""
        p = PacketDyscomSys(sys_type, filename)
        ack: PacketDyscomSysAck = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomSys")
        logger().info("Dyscom sys, type: %s, state: %s, filename: %s", ack.sys_type.name, ack.state.name, ack.filename)
        return DyscomSysResult(ack.sys_type, ack.state, ack.filename)


    def send_send_file_ack(self, block_number: int):
        """Sends get dyscom send file ack and returns immediately without waiting for response"""
        p = PacketDyscomSendFileAck(block_number)
        self._send_packet(p)


    async def start(self):
        """Sends get dyscom start and waits for response"""
        p = PacketDyscomStart()
        ack: PacketDyscomStartAck = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomStart")
        logger().info("Dyscom start")


    async def stop(self):
        """Sends get dyscom stop and waits for response"""
        p = PacketDyscomStop()
        ack: PacketDyscomStopAck = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "DyscomStop")
        logger().info("Dyscom stop")



"""Provides low level types"""

from enum import IntEnum


class LowLevelResult(IntEnum):
    """Represent low level type for result"""
    SUCCESSFUL = 0
    TRANSFER_ERROR = 1
    PARAMETER_ERROR = 2
    TIMEOUT_STIMULATION = 3


class LowLevelMode(IntEnum):
    """Represent low level type for measurement"""
    NO_MEASUREMENT = 0
    STIM_CURRENT = 1
    STIM_VOLTAGE = 2
    HIGH_VOLTAGE_SOURCE = 3


class LowLevelHighVoltageSource(IntEnum):
    """Represent low level high voltage source"""
    STANDARD = 0
    OFF = 1



"""Provides a packet factory class"""

from science_mode_4.utils.logger import logger
from .packet import Packet, PacketAck


class PacketFactory():
    """Packet factory class, it is possible to register multiple packet classes
    per command distinguished by kind"""


    def __init__(self):
        # keys for dict: command, kind, packet class
        self.data: dict[int, int, Packet] = {}
        # register all subclasses of Packet (exclude Packet and PacketAck, because these are base classes)
        self._handle_class(Packet)


    def register_packet(self, packet: Packet):
        """Register a packet"""
        logger().debug("Register packet: %s", packet.__class__.__name__)
        self.data[packet.command, packet.kind] = packet


    def create_packet(self, command: int) -> Packet:
        """Create a packet based on command number"""
        copy = self.data[command, -1].create_copy()
        return copy


    def create_packet_with_data(self, command: int, number: int, data: bytes) -> PacketAck:
        """Create a acknowledge packet based on command number with data"""
        # we use default kind of -1 to have a packet class who is able to read kind from data
        proto: PacketAck = self.data[command, -1]
        copy = proto.create_copy_with_data(data)
        # check if we have a specialized kind
        kind = copy.get_kind(data)
        if kind != -1:
            proto = self.data[command, kind]
            copy = proto.create_copy_with_data(data)
        copy.number = number
        logger().debug("Read package, %s", copy)
        return copy


    def _handle_class(self, cls: type[Packet]):
        """Register all subclasses from cls"""
        for x in cls.__subclasses__():
            # ignore PacketAck
            if x is not PacketAck:
                if issubclass(x, PacketAck):
                    self.register_packet(x(None))
                else:
                    self.register_packet(x())

            # handle subclasses of this class
            self._handle_class(x)



"""Helper class for dyscom"""

import struct
import datetime
from science_mode_4.utils.byte_builder import ByteBuilder


class DyscomHelper:
    """Provides some helper functions"""

    _unpack_func = struct.Struct("<BBBBBBBHh").unpack

    @staticmethod
    def datetime_to_bytes(dt: datetime.datetime) -> bytes:
        """Converts a datetime into dyscom compatible bytes"""
        bb = ByteBuilder()
        bb.append_byte(dt.hour)
        dst = dt.dst()
        bb.append_byte(0 if dst is None or dst == 0 else 1)
        bb.append_byte(dt.day)
        bb.append_byte(dt.minute)
        bb.append_byte(dt.month)
        bb.append_byte(dt.second)
        bb.append_byte((dt.weekday() + 1) % 7)
        bb.append_value(dt.timetuple().tm_yday - 1, 2, True)
        bb.append_value(dt.year - 1900, 2, True)

        return bb.get_bytes()


    @staticmethod
    def bytes_to_datetime(data: bytes) -> datetime.datetime:
        """Converts dyscom datetime bytes to a datetime"""
        # pylint:disable=unused-variable
        hour, dst, day, minute, month, second, _, _, year_since_1900 = \
            DyscomHelper._unpack_func(data)

        # do we need to consider dst?
        result = datetime.datetime(1900 + year_since_1900, month, day, hour, minute, second)
        return result


    @staticmethod
    def str_to_bytes(value: str, byte_count: int) -> bytes:
        """Converts value to bytes with byte_count bytes, last byte will always be 0"""
        temp = bytearray(value.zfill(byte_count), "ascii")
        for x in range(len(value), byte_count):
            temp[x] = 0
        temp[byte_count-1] = 0
        return bytes(temp)


    @staticmethod
    def bytes_to_str(value: bytes, byte_count: int) -> str:
        """Converts bytes to str with byte_count bytes"""
        end = value.find(bytes([0]))
        if end == -1:
            end = byte_count
        return value[0:end].decode()



"""Provides class for ADS129x chip config register 3"""

from dataclasses import dataclass
from enum import IntEnum


class Ads129xPowerDownReferenceBuffer(IntEnum):
    """Represent ADS129xtype for power down reference buffer (configuration register 3, 0x03)"""
    DISABLE_INTERNAL_REFERENCE_BUFFER = 0
    ENABLE_INTERNAL_REFERENCE_BUFFER = 1


class Ads129xReferenceVoltage(IntEnum):
    """Represent ADS129xtype for reference voltage (configuration register 3, 0x03)"""
    VREF_2_4 = 0
    VREF_4_0 = 1


class Ads129xRldMeasurement(IntEnum):
    """Represent ADS129xtype for RLD measurement (configuration register 3, 0x03)"""
    OPEN = 0
    ROUTED = 1


class Ads129xRldReferenceSignal(IntEnum):
    """Represent ADS129xtype for RLD reference signal (configuration register 3, 0x03)"""
    GENERATED_EXTERNALLY = 0
    GENERATED_INTERNALLY = 1


class Ads129xRldBufferPower(IntEnum):
    """Represent ADS129xtype for RLD buffer power (configuration register 3, 0x03)"""
    BUFFER_POWERED_DOWN = 0
    BUFFER_ENABLED = 1


class Ads129xRldSenseFunction(IntEnum):
    """Represent ADS129xtype for RLD sense function (configuration register 3, 0x03)"""
    SENSE_DISABLED = 0
    SENSE_ENABLED = 1


class Ads129xRldLeadOffStatus(IntEnum):
    """Represent ADS129xtype for RLD lead off status (configuration register 3, 0x03)"""
    CONNECTED = 0
    DISCONNECTED = 1


@dataclass
class Ads129xConfigRegister3:
    """Describes config register 3 of ADS129x chip"""

    power_down_reference_buffer = Ads129xPowerDownReferenceBuffer.ENABLE_INTERNAL_REFERENCE_BUFFER
    reference_voltage = Ads129xReferenceVoltage.VREF_4_0
    rld_measurement = Ads129xRldMeasurement.ROUTED
    rld_reference_signal = Ads129xRldReferenceSignal.GENERATED_INTERNALLY
    rld_buffer_power = Ads129xRldBufferPower.BUFFER_ENABLED
    rld_sense_function = Ads129xRldSenseFunction.SENSE_DISABLED
    rld_lead_off_status = Ads129xRldLeadOffStatus.CONNECTED


    def set_data(self, data: bytes):
        """Convert data to information"""
        tmp = data[0]
        self.power_down_reference_buffer = Ads129xPowerDownReferenceBuffer((tmp >> 7) & 0x01)
        self.reference_voltage = Ads129xReferenceVoltage((tmp >> 5) & 0x01)
        self.rld_measurement = Ads129xRldMeasurement((tmp >> 4) & 0x01)
        self.rld_reference_signal = Ads129xRldReferenceSignal((tmp >> 4) & 0x01)
        self.rld_buffer_power = Ads129xRldBufferPower((tmp >> 3) & 0x01)
        self.rld_sense_function = Ads129xRldSenseFunction((tmp >> 1) & 0x01)
        self.rld_lead_off_status = Ads129xRldLeadOffStatus((tmp >> 0) & 0x01)


    def get_data(self) -> bytes:
        """Convert information to bytes"""
        tmp = ((self.power_down_reference_buffer << 7) | (1 << 6) |
               (self.reference_voltage << 5) |(self.rld_measurement << 4) |
               (self.rld_reference_signal << 3) | (self.rld_buffer_power << 2) |
               (self.rld_sense_function << 1) | (self.rld_lead_off_status << 0))
        return [tmp]



"""Provides packet classes for mid level GetCurrentData"""

from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.packet import Packet, PacketAck
from science_mode_4.protocol.types import ResultAndError
from science_mode_4.utils.bit_vector import BitVector
from science_mode_4.utils.byte_builder import ByteBuilder


class PacketMidLevelGetCurrentData(Packet):
    """Packet for mid level GetCurrentData"""

    def __init__(self):
        super().__init__()
        self._command = Commands.MID_LEVEL_GET_CURRENT_DATA


    def get_data(self) -> bytes:
        bb = ByteBuilder()
        bb.append_byte(4)
        return bb.get_bytes()


class PacketMidLevelGetCurrentDataAck(PacketAck):
    """Packet for mid level GetCurrentData acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.MID_LEVEL_GET_CURRENT_DATA_ACK
        self._result_error = ResultAndError.NO_ERROR
        self._data_selection = 0
        self._is_stimulation_active_per_channel = [False] * 8
        self._channel_error = [ResultAndError.NO_ERROR] * 8

        if not data is None:
            self._result_error = ResultAndError(data[0])
            # data selection should always be 4 (Returns all error from all channels)
            self._data_selection = data[1]
            self._is_stimulation_active_per_channel = [x != 0 for x in BitVector.init_from_int(data[2], 8)]
            bb = ByteBuilder()
            bb.append_bytes(data[3:7])
            for index, _ in enumerate(self._channel_error):
                tmp = bb.get_bit_from_position(index * 4, 4)
                if tmp == 0:
                    self._channel_error[index] = ResultAndError.NO_ERROR
                elif tmp == 1:
                    self._channel_error[index] = ResultAndError.ELECTRODE_ERROR
                elif tmp == 2:
                    self._channel_error[index] = ResultAndError.PULSE_TIMEOUT_ERROR
                elif tmp == 3:
                    self._channel_error[index] = ResultAndError.PULSE_LOW_CURRENT_ERROR


    @property
    def result_error(self) -> ResultAndError:
        """Getter for ResultError"""
        return self._result_error


    @property
    def is_stimulation_active_per_channel(self) -> list[bool]:
        """Getter for IsStimulationActive"""
        return self._is_stimulation_active_per_channel


    @property
    def channel_error(self) -> list[int]:
        """Getter for ChannelError"""
        return self._channel_error



"""Provides a simple BitVector class"""


class BitVector():
    """Simple bitvector class"""


    @staticmethod
    def init_from_int(value: int = 0, bit_length: int = 0) -> "BitVector":
        """Creates new BitVector instance from an integer value with bit_length"""
        result = BitVector()
        result.set_from_int(value, bit_length)
        return result


    def __init__(self):
        self._data: list[int] = []
        self._current = 0

        self.set_from_int(0, 0)


    def set_from_int(self, value: int = 0, bit_length: int = 0):
        """Set to an integer value with bit_length"""
        bl = bit_length
        if bl == 0:
            bl = value.bit_length()
        self._data = [0] * bl

        for x in range(bl):
            self._data[x] = (value >> x) & 0x1


    def __getitem__(self, index: int) -> int:
        if (index < 0) or (index >= len(self._data)):
            raise ValueError(f"Bit vector index out of bounds {index} [0 - {len(self._data)}]")

        return self._data[index]


    def __setitem__(self, index: int, value: int):
        if not value in {0, 1}:
            raise ValueError(f"Bit vector wrong value {value}")
        if (index < 0) or (index >= len(self._data)):
            raise ValueError(f"Bit vector index out of bounds {index} [0 - {len(self._data)}]")

        self._data[index] = value


    def __len__(self) -> int:
        return len(self._data)


    def __iter__(self):
        yield from self._data


    def set_length(self, new_length: int):
        """Set length to new_length, does preserve current data"""
        length_difference = new_length - len(self._data)
        if length_difference > 0:
            self.extend(BitVector.init_from_int(0, length_difference))
        elif length_difference < 0:
            self._data = self._data[0:new_length]


    def extend(self, value: "BitVector"):
        """Extends current data with value"""
        if isinstance(value, BitVector):
            self._data += value._data # pylint: disable=protected-access


    def get_bytes(self) -> bytes:
        """Convert to bytes"""
        result = bytearray()
        position = 0

        value = 0
        bit_counter = 0
        bl = len(self._data)
        while position < bl:
            if bit_counter == 8:
                result.append(value)
                value = 0
                bit_counter = 0

            value |= self[position] << bit_counter
            position += 1
            bit_counter += 1

        if bit_counter > 0:
            result.append(value)

        return bytes(result)


    def __repr__(self) -> str:
        return f"{type(self).__name__}(0b{self._data: _b})"


    def __str__(self) -> str:
        return "0b" + format(self._data, "_b")



"""Provides an example how to use low level layer to measure current and plot values"""

import asyncio
import sys

import matplotlib.pyplot as plt
import numpy as np

from science_mode_4 import DeviceP24
from science_mode_4 import ChannelPoint, Commands
from science_mode_4 import SerialPortConnection
from science_mode_4 import PacketLowLevelChannelConfigAck
from science_mode_4 import Connector, Channel
from science_mode_4 import LowLevelHighVoltageSource, LowLevelMode
from examples.utils.example_utils import ExampleUtils


async def main() -> int:
    """Main function"""

    # get comport from command line argument
    com_port = ExampleUtils.get_comport_from_commandline_argument()
    # create serial port connection
    connection = SerialPortConnection(com_port)
    # open connection, now we can read and write data
    connection.open()

    # create science mode device
    device = DeviceP24(connection)
    # call initialize to get basic information (serial, versions) and stop any active stimulation/measurement
    # to have a defined state
    await device.initialize()

    # get low level layer to call low level commands
    low_level_layer = device.get_layer_low_level()

    # call init low level and enable measurement
    await low_level_layer.init(LowLevelMode.STIM_CURRENT, LowLevelHighVoltageSource.STANDARD)

    # send one channel config so we get only one acknowledge
    low_level_layer.send_channel_config(True, Channel.RED, Connector.GREEN,
                                        [ChannelPoint(1000, 40), ChannelPoint(1000, 0),
                                        ChannelPoint(1000, -20)])

    # wait for stimulation to happen
    await asyncio.sleep(0.1)

    measurement_sample_time = 0
    measurement_samples: list[float] = []
    # get new data from connection
    ack = low_level_layer.packet_buffer.get_packet_from_buffer()
    if ack:
        if ack.command == Commands.LOW_LEVEL_CHANNEL_CONFIG_ACK:
            ll_config_ack: PacketLowLevelChannelConfigAck = ack
            measurement_sample_time = ll_config_ack.sampling_time_in_microseconds
            measurement_samples.extend(ll_config_ack.measurement_samples)

    # call stop low level
    await low_level_layer.stop()

    # close serial port connection
    connection.close()

    # show plot
    _, ax = plt.subplots()
    # use measurement_sample_time as time frame for x-axis
    # this may be wrong because measurement_sample_time seems to long and
    # does not match actual stimulation duration
    ax.plot(np.linspace(0, measurement_sample_time, len(measurement_samples)),
            measurement_samples)

    ax.set(xlabel="Sample Time (µs)", ylabel="Current (mA)",
        title="Current measurement")
    ax.grid()

    plt.show()

    return 0


if __name__ == "__main__":
    res = asyncio.run(main())
    sys.exit(res)



"""Provides class for ADS129x chip respiration control register"""

from dataclasses import dataclass
from enum import IntEnum


class Ads129xRespirationDemodulationCircuitry(IntEnum):
    """Represent ADS129xtype for respiration demodulation circuitry (respiration control register, 0x16)"""
    DEMODULATION_CIRCUITRY_OFF = 0
    DEMODULATION_CIRCUITRY_ON = 1


class Ads129xRespirationModulationCircuitry(IntEnum):
    """Represent ADS129xtype for respiration modulation circuitry (respiration control register, 0x16)"""
    MODULATION_CIRCUITRY_OFF = 0
    MODULATION_CIRCUITRY_ON = 1


class Ads129xRespirationPhase(IntEnum):
    """Represent ADS129xtype for respiration phase (respiration control register, 0x16)"""
    PHASE_22_5 = 0
    PHASE_45_0 = 1
    PHASE_67_5 = 2
    PHASE_90_0 = 3
    PHASE_112_5 = 4
    PHASE_135_0 = 5
    PHASE_157_5 = 6


class Ads129xRespirationControl(IntEnum):
    """Represent ADS129xtype for respiration control (respiration control register, 0x16)"""
    NO_RESPIRATION = 0
    EXTERNAL_RESPIRATION = 1
    INTERNAL_RESPIRATION_WITH_INTERNAL_SIGNAL = 2
    INTERNAL_RESPIRATION_WITH_USER_GENERATED_SIGNAL = 3


@dataclass
class Ads129xRespirationControlRegister:
    """Describes respiration control register of ADS129x chip"""

    respiration_demodulation_circuitry = Ads129xRespirationDemodulationCircuitry.DEMODULATION_CIRCUITRY_ON
    respiration_modulation_circuitry = Ads129xRespirationModulationCircuitry.MODULATION_CIRCUITRY_ON
    respiration_phase = Ads129xRespirationPhase.PHASE_67_5
    respiration_control = Ads129xRespirationControl.INTERNAL_RESPIRATION_WITH_INTERNAL_SIGNAL


    def set_data(self, data: bytes):
        """Convert data to information"""
        tmp = data[0]
        self.respiration_demodulation_circuitry = Ads129xRespirationDemodulationCircuitry((tmp >> 7) & 0x01)
        self.respiration_modulation_circuitry = Ads129xRespirationModulationCircuitry((tmp >> 6) & 0x01)
        self.respiration_phase = Ads129xRespirationPhase((tmp >> 2) & 0x07)
        self.respiration_control = Ads129xRespirationControl((tmp >> 0) & 0x03)


    def get_data(self) -> bytes:
        """Convert information to bytes"""
        tmp = ((self.respiration_demodulation_circuitry << 7) |
               (self.respiration_modulation_circuitry << 6) | (1 << 5) |
               (self.respiration_phase << 2) | (self.respiration_control << 0))
        return [tmp]



"""Provides common types"""

from enum import IntEnum


class Channel(IntEnum):
    """Represents a channel"""
    RED = 0
    BLUE = 1
    BLACK = 2
    WHITE = 3


class Connector(IntEnum):
    """Represents a cable connector"""
    YELLOW = 0
    GREEN = 1


class ResultAndError(IntEnum):
    """Represents science mode type ResultAndError"""
    NO_ERROR = 0
    TRANSFER_ERROR = 1
    PARAMETER_ERROR = 2
    NOT_INITIALIZED = 7
    ELECTRODE_ERROR = 10
    PULSE_TIMEOUT_ERROR = 16
    PULSE_LOW_CURRENT_ERROR = 28


class StimStatus(IntEnum):
    """Represents science mode type StimStatus"""
    NO_LEVEL_INITIALIZED = 0
    LOW_LEVEL_INITIALIZED = 1
    MID_LEVEL_INITIALIZED = 2
    MID_LEVEL_RUNNING = 3



"""Provides classes for general GetVersion"""

from typing import NamedTuple
from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.packet import Packet, PacketAck


class GetExtendedVersionResult(NamedTuple):
    """Helper class for dyscom get with type file system status"""
    firmware_version: str
    science_mode_version: str


class PacketGeneralGetExtendedVersion(Packet):
    """Packet for general GetExtendetVersion"""


    def __init__(self):
        super().__init__()
        self._command = Commands.GET_EXTENDED_VERSION


class PacketGeneralGetExtendedVersionAck(PacketAck):
    """Packet for general GetExtendetVersion acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.GET_EXTENDED_VERSION_ACK
        self._successful = False
        self._firmware_version = ""
        self._science_mode_version = ""
        self._firmware_hash = 0
        self._hash_type = 0
        self._is_valid_hash = False

        if not data is None:
            self._successful = data[0] == 0
            self._firmware_version = f"{data[1]}.{data[2]}.{data[3]}"
            self._science_mode_version = f"{data[4]}.{data[5]}.{data[6]}"
            self._firmware_hash = int.from_bytes(data[7:10], "little")
            self._hash_type = data[11]
            self._is_valid_hash = data[12] == 1


    @property
    def successful(self) -> bool:
        """Getter for Successful"""
        return self._successful


    @property
    def firmware_version(self) -> str:
        """Getter for FirmwareVersion"""
        return self._firmware_version


    @property
    def science_mode_version(self) -> str:
        """Getter for ^ScienceModeVersion"""
        return self._science_mode_version


    @property
    def firmware_hash(self) -> int:
        """Getter for FirmwareHash"""
        return self._firmware_hash


    @property
    def hash_type(self) -> int:
        """Getter for HashType"""
        return self._hash_type


    @property
    def is_valid_hash(self) -> bool:
        """Getter for IsValidHash"""
        return self._is_valid_hash



"""Provides a buffer class to plot values"""

from queue import Empty, Full, Queue
import matplotlib.pyplot as plt
import matplotlib.animation as animation

from examples.utils.plot_base import PlotHelper, PlotValueChannel


class PyPlotValueChannel(PlotValueChannel):
    """Class for holding a specific number of values and handling minimum/maximum specific for PyPlot"""


    def __init__(self, axes: plt.Axes, max_value_count: int, color: str):

        super().__init__(max_value_count)

        # add line with color
        line, = axes.plot(self._x_data, self._y_data, color)
        # enable grid
        axes.grid()

        self._axes = axes
        self._line = line

        self._minimum = None
        self._maximum = None

        # this queue is used to synchronize data between background and main thread
        self._data_queue = Queue(maxsize=1)


    def append_value(self, value: float):
        """Appends a value and updates minimum and maximum for plot. Does run in main thread context"""
        super().append_value(value)

        if self._minimum is None or value < self._minimum:
            self._minimum = value
        if self._maximum is None or value > self._maximum:
            self._maximum = value

        # self._axes.redraw_in_frame()
        try:
            self._data_queue.put_nowait([self._x_data, self._y_data, self._minimum, self._maximum])
        except Full:
            # Queue is full, skip this update
            pass


    def update_plot(self):
        """This function is call in context animation function"""
        super().update_plot()
        try:
            new_x_data, new_y_data, new_minimum, new_maximum = self._data_queue.get_nowait()
            self._line.set_xdata(new_x_data)
            self._line.set_ydata(new_y_data)

            if len(new_x_data) > 1:
                self._axes.set_xlim(new_x_data[0], new_x_data[-1])
                offset = (new_maximum - new_minimum) * 0.1
                self._axes.set_ylim(bottom=new_minimum - offset, top=new_maximum + offset)
        except Empty:
            # No new data in the queue
            pass


class PyPlotHelper(PlotHelper):
    """Class to handle values for plotting"""


    def __init__(self, channels: dict[int, tuple[str, str]], max_value_count: int):
        super().__init__()

        x_dimension, y_dimension = self._calc_layout_dimension(len(channels))
        self._figure, self._axes = plt.subplots(x_dimension, y_dimension, constrained_layout=True, squeeze=False)

        sub_plot_counter = 0

        for key, value in channels.items():
            x_pos, y_pos = self._calc_layout_pos(sub_plot_counter)
            ax = self._axes[x_pos, y_pos]
            sub_plot_counter += 1

            ax.set(xlabel="Samples", ylabel=value[0], title=value[0])
            self._data[key] = PyPlotValueChannel(ax, max_value_count, value[1])

        # interactive mode and show plot
        self._animation_result = animation.FuncAnimation(self._figure, self._animation, interval=100)
        # plt.ion()
        plt.show(block=False)


    def update(self):
        plt.pause(0.0001)


    def _animation(self, frame: int, *fargs: tuple): # pylint:disable=unused-argument
        """This function is call in context of main thread"""
        for x in self._data.values():
            x.update_plot()



"""Provides base class for all ScienceMode layers"""

from .protocol.protocol_helper import ProtocolHelper
from .protocol.types import ResultAndError
from .protocol.packet import Packet, PacketAck
from .protocol.packet_factory import PacketFactory
from .protocol.packet_number_generator import PacketNumberGenerator
from .utils.packet_buffer import PacketBuffer


class Layer():
    """Base class for all layers"""


    def __init__(self, packet_buffer: PacketBuffer, packet_factory: PacketFactory, packet_number_generator: PacketNumberGenerator):
        self._packet_factory = packet_factory
        self._packet_number_generator = packet_number_generator
        self._packet_buffer = packet_buffer


    @property
    def packet_buffer(self) -> PacketBuffer:
        """Getter for packet buffer"""
        return self._packet_buffer


    def _send_packet(self, packet: Packet):
        """Generates a new packet number and send packet"""
        ack = ProtocolHelper.send_packet(packet, self._packet_number_generator.get_next_number(),
                                         self._packet_buffer)
        return ack


    async def _send_packet_and_wait(self, packet: Packet) -> PacketAck:
        """Generates a new packet number, send packet and waits for response"""
        ack = await ProtocolHelper.send_packet_and_wait(packet, self._packet_number_generator.get_next_number(),
                                                        self._packet_buffer)
        return ack


    def _check_result_error(self, result_error: ResultAndError, packet_name: str):
        """Check if result_error contains an error and if yes prints packet_name"""
        if result_error != ResultAndError.NO_ERROR:
            raise ValueError(f"Error {packet_name} {result_error.name}")



"""Provides class for ADS129x chip"""

from dataclasses import dataclass

from science_mode_4.utils.byte_builder import ByteBuilder
from .ads129x_config_register_1 import Ads129xConfigRegister1
from .ads129x_config_register_2 import Ads129xConfigRegister2
from .ads129x_config_register_3 import Ads129xConfigRegister3
from .ads129x_config_register_4 import Ads129xConfigRegister4
from .ads129x_respiration_control_register import Ads129xRespirationControlRegister


@dataclass
class Ads129x:
    """Describes register map of ADS129x chip"""

    # control register
    device_id = 0

    # config register
    config_register_1 = Ads129xConfigRegister1() # CONFIG1
    config_register_2 = Ads129xConfigRegister2() # CONFIG2
    config_register_3 = Ads129xConfigRegister3() # CONFIG3
    config_register_4 = Ads129xConfigRegister4() # CONFIG4

    # signal derivation register
    positive_signal_derivation_register = 0x02 # RLD_SENSP
    negative_signal_derivation_register = 0xEA # RLD_SENSN

    # respiration register
    respiration_control_register = Ads129xRespirationControlRegister() # RESP


    def set_data(self, data: bytes):
        """Convert data to information"""

        control_register = data[0]
        self.device_id = control_register

        self.config_register_1.set_data([data[8]])
        self.config_register_2.set_data([data[9]])
        self.config_register_3.set_data([data[10]])
        self.config_register_4.set_data([data[11]])

        self.positive_signal_derivation_register = data[23]
        self.negative_signal_derivation_register = data[22]

        self.respiration_control_register.set_data([data[21]])


    def get_data(self) -> bytes:
        """Convert information to bytes"""

        bb = ByteBuilder(0, 26)
        bb.set_bytes_to_position(self.config_register_1.get_data(), 8, 1)
        bb.set_bytes_to_position(self.config_register_2.get_data(), 9, 1)
        bb.set_bytes_to_position(self.config_register_3.get_data(), 10, 1)
        bb.set_bytes_to_position(self.config_register_4.get_data(), 11, 1)

        bb.set_bytes_to_position([self.positive_signal_derivation_register], 23, 1)
        bb.set_bytes_to_position([self.negative_signal_derivation_register], 22, 1)

        bb.set_bytes_to_position(self.respiration_control_register.get_data(), 21, 1)

        return bb.get_bytes()



"""Provides packet classes for dyscom get with type operation mode"""

from science_mode_4.protocol.commands import Commands
from .dyscom_types import DyscomGetType, DyscomGetOperationModeType
from .dyscom_get import PacketDyscomGet, PacketDyscomGetAck


class PacketDyscomGetOperationMode(PacketDyscomGet):
    """Packet for dyscom get with type operation mode"""


    def __init__(self):
        super().__init__()
        self._command = Commands.DL_GET
        self._type = DyscomGetType.OPERATION_MODE
        self._kind = int(self._type)


class PacketDyscomGetAckOperationMode(PacketDyscomGetAck):
    """Packet for dyscom get for type operation mode acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._kind = int(DyscomGetType.OPERATION_MODE)
        self._operation_mode = DyscomGetOperationModeType.UNDEFINED

        if not data is None:
            self._operation_mode = DyscomGetOperationModeType(data[2])


    @property
    def operation_mode(self) -> DyscomGetOperationModeType:
        """Getter for operation mode"""
        return self._operation_mode



"""Provides packet classes for dyscom get with type list of measurement meta info"""

from science_mode_4.protocol.commands import Commands
from .dyscom_types import DyscomGetType
from .dyscom_get import PacketDyscomGet, PacketDyscomGetAck


class PacketDyscomGetListOfMeasurementMetaInfo(PacketDyscomGet):
    """Packet for dyscom get with type list of measurement meta info"""


    def __init__(self):
        super().__init__()
        self._command = Commands.DL_GET
        self._type = DyscomGetType.LIST_OF_MEASUREMENT_META_INFO
        self._kind = int(self._type)


class PacketDyscomGetAckListOfMeasurementMetaInfo(PacketDyscomGetAck):
    """Packet for dyscom get for type list of measurement meta info acknowledge"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._kind = int(DyscomGetType.LIST_OF_MEASUREMENT_META_INFO)
        self._number_of_measurements = 0

        if not data is None:
            self._number_of_measurements = int.from_bytes(data[2:4], "little")


    @property
    def number_of_measurements(self) -> int:
        """Getter for number of measurements"""
        return self._number_of_measurements



"""Init file for protocol"""

from .channel_point import *
from .commands import *
from .packet_factory import *
from .packet_number_generator import *
from .packet import *
from .protocol import *
# from .protocol_helper import *
from .types import *



"""Provides a class for a serial connection"""

import serial
import serial.tools.list_ports
import serial.tools.list_ports_common

from .connection import Connection
from .logger import logger


class SerialPortConnection(Connection):
    """Serial connection class"""


    @staticmethod
    def list_ports() -> list[serial.tools.list_ports_common.ListPortInfo]:
        """Returns a list of all serial ports"""
        return serial.tools.list_ports.comports()


    @staticmethod
    def list_science_mode_device_ports() -> list[serial.tools.list_ports_common.ListPortInfo]:
        """Returns a list of all serial ports with a science mode device"""
        ports = SerialPortConnection.list_ports()
        # science mode devices (P24/I24) have an STM32 mcu and these are
        # default values for USB CDC devices
        filtered_ports = list(filter(lambda x: x.vid == 0x0483 and x.pid == 0x5740, ports))
        return filtered_ports


    def __init__(self, port: str):
        self._ser = serial.Serial(timeout = 0)
        self._ser.port = port


    def open(self):
        self._ser.open()
        self._ser.set_buffer_size(4096*128)


    def close(self):
        self._ser.close()


    def is_open(self) -> bool:
        return self._ser.is_open


    def write(self, data: bytes):
        self._ser.write(data)


    def read(self) -> bytes:
        result = []
        if self._ser.in_waiting > 0:
            result = self._ser.read_all()
            logger().debug("Incoming data, length: %d, bytes: %s", len(result), result.hex(" ").upper())
        return bytes(result)


    def clear_buffer(self):
        self._ser.reset_input_buffer()



"""Test program how to use library without installing the library,
DO NOT USE THIS FILE, USE EXAMPLES INSTEAD"""

import sys
import asyncio

from science_mode_4.device_p24 import DeviceP24
from science_mode_4.utils.serial_port_connection import SerialPortConnection



async def main() -> int:
    """Main function"""

    devices = SerialPortConnection.list_science_mode_device_ports()
    connection = SerialPortConnection(devices[0].device)
    # devices = UsbConnection.list_science_mode_devices()
    # connection = UsbConnection(devices[0])
    # connection = NullConnection()
    connection.open()

    device = DeviceP24(connection)
    await device.initialize()
    general = device.get_layer_general()
    print(f"Device id: {general.device_id}")
    print(f"Firmware version: {general.firmware_version}")
    print(f"Science mode version: {general.science_mode_version}")

    connection.close()

    return 0


if __name__ == "__main__":
    res = asyncio.run(main())
    sys.exit(res)



"""Init file for dyscom"""

from .dyscom_get_battery_status import *
from .dyscom_get_device_id import *
from .dyscom_get_file_by_name import *
from .dyscom_get_file_info import *
from .dyscom_get_file_system_status import *
from .dyscom_get_firmware_version import *
from .dyscom_get_list_of_measurement_meta_info import *
from .dyscom_get_operation_mode import *
from .dyscom_init import *
from .dyscom_layer import *
from .dyscom_power_module import *
from .dyscom_send_file import *
from .dyscom_send_live_data import *
from .dyscom_send_measurement_meta_info import *
from .dyscom_start import *
from .dyscom_stop import *
from .dyscom_sys import *
from .dyscom_types import *
from .ads129x import *



"""Provides packet classes for dyscom send measurement meta info"""

import datetime
from science_mode_4.protocol.commands import Commands
from science_mode_4.protocol.packet import PacketAck
from .dyscom_helper import DyscomHelper
from .dyscom_types import DyscomInitParams


class PacketDyscomSendMeasurementMetaInfo(PacketAck):
    """Packet for dyscom send measurement meta info (this is technically not an acknowledge, but it is handled as such,
    because it is send automatically from device)"""


    def __init__(self, data: bytes):
        super().__init__(data)
        self._command = Commands.DL_SEND_FILE
        self._init_params: DyscomInitParams = DyscomInitParams()
        self._file_name = ""
        self._file_size = 0
        self._file_number = 0
        self._proband_name = ""
        self._start_time = datetime.datetime.now()
        self._duration = datetime.timedelta()

        if not data is None:
            self._init_params.set_data(data[0:361])
            self._file_name = DyscomHelper.bytes_to_str(data[361:421], 60)
            self._file_size = int.from_bytes(data[421:429], "big")
            self._file_number = int.from_bytes(data[429:431], "big")
            self._proband_name = DyscomHelper.bytes_to_str(data[431:468], 37)
            self._start_time = DyscomHelper.bytes_to_datetime(data[468:479])
            self._duration = datetime.timedelta(seconds=int.from_bytes(data[479:483]))


    @property
    def init_params(self) -> DyscomInitParams:
        """Getter for init params"""
        return self._init_params


    @property
    def file_name(self) -> str:
        """Getter for file id"""
        return self._file_name


    @property
    def file_size(self) -> int:
        """Getter for file size"""
        return self._file_size


    @property
    def file_number(self) -> int:
        """Getter for file number"""
        return self._file_number


    @property
    def proband_name(self) -> str:
        """Getter for proband name"""
        return self._proband_name


    @property
    def start_time(self) -> datetime.datetime:
        """Getter for start time"""
        return self._start_time


    @property
    def duration(self) -> datetime.timedelta:
        """Getter for duration"""
        return self._duration



"""Example how to use dyscom with fastplotlib for fast plotting of values.
Because fastplotlib requires its own event loop, we handle all device communication in a
background thread and synchronize measurement data vie Queue to main thread"""

import asyncio
import threading

import fastplotlib as fpl

from science_mode_4 import DeviceI24
from science_mode_4 import Commands
from science_mode_4 import SerialPortConnection
from science_mode_4.dyscom.ads129x.ads129x_config_register_1 import Ads129xOutputDataRate, Ads129xPowerMode
from science_mode_4.dyscom.dyscom_get_operation_mode import PacketDyscomGetAckOperationMode
from science_mode_4.dyscom.dyscom_send_live_data import PacketDyscomSendLiveData
from science_mode_4.dyscom.dyscom_types import DyscomGetType, DyscomInitParams, DyscomPowerModulePowerType,\
    DyscomPowerModuleType, DyscomSignalType
from science_mode_4.protocol.types import ResultAndError
from examples.utils.fastplotlib_utils import FastPlotLibHelper
from examples.utils.example_utils import ExampleUtils


def main():
    """Main function"""

    # initialize plot helper the handle plot specific things
    plot_helper = FastPlotLibHelper({0: ["BI", "b"], 1: ["EMG1", "r"], 2: ["EMG2", "y"],\
                        3: ["Breathing", "g"], 4: ["Temperature", "w"]}, 1000)
    # flag to indicate
    is_window_open: bool = True

    async def device_communication() -> int:
        """Communication with science mode device"""

        # get comport from command line argument
        com_port = ExampleUtils.get_comport_from_commandline_argument()
        # create serial port connection
        connection = SerialPortConnection(com_port)
        # open connection, now we can read and write data
        connection.open()

        # create science mode device
        device = DeviceI24(connection)
        # call initialize to get basic information (serial, versions) and stop any active stimulation/measurement
        # to have a defined state
        await device.initialize()

        # get dyscom layer to call dyscom level commands
        dyscom = device.get_layer_dyscom()

        # call enable measurement power module for measurement
        await dyscom.power_module(DyscomPowerModuleType.MEASUREMENT, DyscomPowerModulePowerType.SWITCH_ON)
        # call init with lowest sample rate and enable signal types
        init_params = DyscomInitParams()
        init_params.signal_type = [DyscomSignalType.BI, DyscomSignalType.EMG_1,\
                                DyscomSignalType.EMG_2, DyscomSignalType.BREATHING, DyscomSignalType.TEMPERATURE]
        init_params.register_map_ads129x.config_register_1.output_data_rate = Ads129xOutputDataRate.HR_MODE_500_SPS__LP_MODE_250_SPS
        init_params.register_map_ads129x.config_register_1.power_mode = Ads129xPowerMode.LOW_POWER
        await dyscom.init(init_params)

        # start dyscom measurement
        await dyscom.start()

        # loop for some time
        for x in range(1000):
            # check if we closed window
            if not is_window_open:
                break

            # check operation mode from time to time, this function is not waiting for response
            # so we have to handle it by ourself later
            if x % 100 == 0:
                dyscom.send_get_operation_mode()

            live_data_counter = 0
            while True:
                # process all available packages
                ack = dyscom.packet_buffer.get_packet_from_buffer(live_data_counter == 0)
                if ack:
                    # because there are multiple get commands, we need to additionally check kind,
                    # which is always associated DyscomGetType
                    if ack.command == Commands.DL_GET_ACK and ack.kind == DyscomGetType.OPERATION_MODE:
                        om_ack: PacketDyscomGetAckOperationMode = ack
                        print(f"Operation mode {om_ack.operation_mode.name}")
                        # check if measurement is still active
                        if om_ack.result_error != ResultAndError.NO_ERROR:
                            break
                    elif ack.command == Commands.DL_SEND_LIVE_DATA:
                        live_data_counter += 1

                        sld: PacketDyscomSendLiveData = ack
                        if sld.status_error:
                            print(f"SendLiveData status error {sld.samples}")
                            break

                        # these samples match signal types from dl_init command
                        plot_helper.append_value(0, sld.samples[0].value)
                        plot_helper.append_value(1, sld.samples[1].value)
                        plot_helper.append_value(2, sld.samples[2].value)
                        plot_helper.append_value(3, sld.samples[3].value)
                        plot_helper.append_value(4, sld.samples[4].value)

                        # signal type seems to be always 0, so we can't use it
                        # for s in sld.samples:
                        #     ph.append_value(int(s.signal_type), s.value)

                        plot_helper.update()
                else:
                    # print(f"Live data acknowledges per iteration {live_data_counter}")
                    break

            await asyncio.sleep(0.01)

        # stop measurement
        await dyscom.stop()
        # turn power module off
        await dyscom.power_module(DyscomPowerModuleType.MEASUREMENT, DyscomPowerModulePowerType.SWITCH_OFF)

        # close serial port connection
        connection.close()

        print("Close plot window to end example")

        return 0


    def data_generator() -> int:
        """Background function"""
        return asyncio.run(device_communication())


    # Create and start the data generator thread (aka background thread)
    data_thread = threading.Thread(target=data_generator, daemon=True)
    data_thread.start()

    # fastplotlib requires a main loop to draw everything
    fpl.loop.run()
    is_window_open = False


if __name__ == "__main__":
    main()



"""Provides device class representing a science mode device"""

from enum import IntEnum
from typing import Type

from .layer import Layer
from .protocol.types import StimStatus
from .protocol.packet_factory import PacketFactory
from .protocol.packet_number_generator import PacketNumberGenerator
from .general.general_layer import LayerGeneral
from .low_level.low_level_layer import LayerLowLevel
from .mid_level.mid_level_layer import LayerMidLevel
from .dyscom.dyscom_layer import LayerDyscom
from .dyscom.dyscom_types import DyscomGetOperationModeType
from .utils.connection import Connection
from .utils.packet_buffer import PacketBuffer


class DeviceCapability(IntEnum):
    """Represent device capabilities"""
    GENERAL = 0
    LOW_LEVEL = 1
    MID_LEVEL = 2
    DYSCOM = 3


class Device():
    """Base class for a science mode devices"""


    def __init__(self, conn: Connection, capabilities: set[DeviceCapability]):
        self._connection  = conn
        self._packet_factory = PacketFactory()
        self._packet_buffer = PacketBuffer(self._connection, self._packet_factory)
        self._packet_number_generator = PacketNumberGenerator()
        self._capabilities = capabilities
        self._layer: dict[DeviceCapability, Layer] = {}

        self._add_layer(DeviceCapability.GENERAL, capabilities, LayerGeneral)
        self._add_layer(DeviceCapability.LOW_LEVEL, capabilities, LayerLowLevel)
        self._add_layer(DeviceCapability.MID_LEVEL, capabilities, LayerMidLevel)
        self._add_layer(DeviceCapability.DYSCOM, capabilities, LayerDyscom)


    @property
    def connection(self) -> Connection:
        """Getter for connection"""
        return self._connection


    @property
    def packet_factory(self) -> PacketFactory:
        """Getter for packet factory"""
        return self._packet_factory


    @property
    def packet_buffer(self) -> PacketBuffer:
        """Getter for packet buffer"""
        return self._packet_buffer


    @property
    def packet_number_generator(self) -> PacketNumberGenerator:
        """Getter for packet number generator"""
        return self._capabilities


    @property
    def capabilities(self) -> set[DeviceCapability]:
        """Getter for capabilities"""
        return self._capabilities


    async def initialize(self):
        """Initialize device to get basic information (serial, versions) and stop any active stimulation/measurement"""
        if {DeviceCapability.LOW_LEVEL, DeviceCapability.MID_LEVEL}.issubset(self._capabilities):
            # get stim status to see if low/mid level is initialized or running
            stim_status = await self.get_layer_general().get_stim_status()
            if stim_status.stim_status == StimStatus.LOW_LEVEL_INITIALIZED:
                await self.get_layer_low_level().stop()
            elif stim_status.stim_status in [StimStatus.MID_LEVEL_INITIALIZED, StimStatus.MID_LEVEL_RUNNING]:
                await self.get_layer_mid_level().stop()
        if DeviceCapability.DYSCOM in self._capabilities:
            # get operation mode to see if dyscom measurement is running
            operation_mode = await self.get_layer_dyscom().get_operation_mode()
            if operation_mode in [DyscomGetOperationModeType.LIVE_MEASURING_PRE,
                                  DyscomGetOperationModeType.LIVE_MEASURING,
                                  DyscomGetOperationModeType.RECORD_PRE,
                                  DyscomGetOperationModeType.RECORD]:
                await self.get_layer_dyscom().stop()

        await self.get_layer_general().initialize()


    def get_layer_general(self) -> LayerGeneral:
        """Helper function to access general layer"""
        return self._layer[DeviceCapability.GENERAL]


    def get_layer_mid_level(self) -> LayerMidLevel:
        """Helper function to access mid level layer"""
        return self._layer[DeviceCapability.MID_LEVEL]


    def get_layer_low_level(self) -> LayerLowLevel:
        """Helper function to access low level layer"""
        return self._layer[DeviceCapability.LOW_LEVEL]


    def get_layer_dyscom(self) -> LayerDyscom:
        """Helper function to access dyscom layer"""
        return self._layer[DeviceCapability.DYSCOM]


    def add_layer(self, capability: DeviceCapability, layer: Layer):
        """Add layer"""
        self._layer[capability] = layer


    def _add_layer(self, capability: DeviceCapability, used_capabilities: set[DeviceCapability], layer_class: Type[Layer]):
        """Helper method that checks if capability is in used_capabilities and if yes add a layer_class instance"""
        if capability in used_capabilities:
            self.add_layer(capability, layer_class(self._packet_buffer, self._packet_factory, self._packet_number_generator))



"""Helper class for sending packets to connection"""

import asyncio

from science_mode_4.general.general_error import PacketGeneralError
from science_mode_4.general.general_unknown_command import PacketGeneralUnknownCommand
from science_mode_4.utils.packet_buffer import PacketBuffer
from .protocol import Protocol
from .commands import Commands
from .packet import Packet, PacketAck


class ProtocolHelper:
    """Helper class for Protocol"""


    @staticmethod
    def send_packet(packet: Packet, packet_number: int, packet_buffer: PacketBuffer) -> PacketAck:
        """Send a packet and returns immediately"""
        packet.number = packet_number
        packet_buffer.add_open_acknowledge(packet)

        packet_buffer.connection.write(Protocol.packet_to_bytes(packet))


    @staticmethod
    async def send_packet_and_wait(packet: Packet, packet_number: int, packet_buffer: PacketBuffer, timeout_in_seconds = 5) -> PacketAck:
        """Send a packet and wait for response, if no response arrives raise an exception,
        this function assumes that the response has the same packet number and ack command must be command+1
        
        Clears all incoming data from connection and packet buffer"""

        # discard all packets because we don"t need them anymore
        packet_buffer.clear_buffer()
        ProtocolHelper.send_packet(packet, packet_number, packet_buffer)

        # calculate number of loops to reach timeout
        sleep_duration = 0.01
        counter = timeout_in_seconds / sleep_duration
        while counter > 0:
            while True:
                ack = packet_buffer.get_packet_from_buffer()
                if ack:
                    if (ack.command == packet.command + 1) and (ack.number == packet.number):
                        return ack

                    # check if we got an error
                    if ack.command == Commands.GENERAL_ERROR:
                        ge: PacketGeneralError = ack
                        raise ValueError(f"General error packet {ge.result_error.name}")
                    if ack.command == Commands.UNKNOWN_COMMAND:
                        uc: PacketGeneralUnknownCommand = ack
                        raise ValueError(f"Unknown command packet {uc.result_error.name}")

                    # discard acknowledge and continue

                # no acknowledge arrived, sleep and check again
                break

            await asyncio.sleep(sleep_duration)
            counter -= 1

        # we got no response in time, so remove open acknowledges
        packet_buffer.remove_open_acknowledge(packet)
        raise ValueError(f"No valid answer for packet {packet.command}")



"""Provides mid level layer"""

from science_mode_4.layer import Layer
from science_mode_4.utils.logger import logger
from .mid_level_current_data import PacketMidLevelGetCurrentData, PacketMidLevelGetCurrentDataAck
from .mid_level_stop import PacketMidLevelStop, PacketMidLevelStopAck
from .mid_level_update import PacketMidLevelUpdate, PacketMidLevelUpdateAck
from .mid_level_types import MidLevelChannelConfiguration
from .mid_level_init import PacketMidLevelInit, PacketMidLevelInitAck


class LayerMidLevel(Layer):
    """Class for mid level layer"""


    async def init(self, do_stop_on_all_errors: bool):
        """Send mid level init command and waits for response"""
        p = PacketMidLevelInit()
        p.do_stop_on_all_errors = do_stop_on_all_errors
        ack: PacketMidLevelInitAck = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "MidLevelInit")
        logger().info("Mid level init")


    async def stop(self):
        """Send mid level stop command and waits for response"""
        p = PacketMidLevelStop()
        ack: PacketMidLevelStopAck = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "MidLevelStop")
        logger().info("Mid level stop")


    async def update(self, channel_configuration: list[MidLevelChannelConfiguration]):
        """Send mid level update command and waits for response"""
        p = PacketMidLevelUpdate()
        p.channel_configuration = channel_configuration
        ack: PacketMidLevelUpdateAck = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "MidLevelUpdate")
        logger().info("Mid level update")


    async def get_current_data(self) -> list[bool]:
        """Send mid level get current data command and waits for response"""
        p = PacketMidLevelGetCurrentData()
        ack: PacketMidLevelGetCurrentDataAck = await self._send_packet_and_wait(p)
        self._check_result_error(ack.result_error, "MidLevelGetCurrentData")
        if True in ack.channel_error:
            raise ValueError(f"Error mid level get current data channel error {ack.channel_error}")
        logger().info("Mid level get current data, active channels: %s", ack.is_stimulation_active_per_channel)
        return ack.is_stimulation_active_per_channel



"""Provides class for ADS129x chip config register 1"""

from dataclasses import dataclass
from enum import IntEnum


class Ads129xPowerMode(IntEnum):
    """Represent ADS129xtype for power mode (configuration register 1, 0x01)"""
    LOW_POWER = 0
    HIGH_RESOLUTION = 1


class Ads129xReadMode(IntEnum):
    """Represent ADS129xtype for read mode (configuration register 1, 0x01)"""
    DAISY_CHAIN = 0
    MULTIPLE_READBACK = 1


class Ads129xClockConnection(IntEnum):
    """Represent ADS129xtype for clock connection (configuration register 1, 0x01)"""
    OSCILLATOR_CLOCK_OUTPUT_DISABLED = 0
    OSCILLATOR_CLOCK_OUTPUT_ENABLED = 1


class Ads129xOutputDataRate(IntEnum):
    """Represent ADS129xtype for output data rate (configuration register 1, 0x01)"""
    HR_MODE_32_KSPS__LP_MODE_16_KSPS = 0
    HR_MODE_16_KSPS__LP_MODE_8_KSPS = 1
    HR_MODE_8_KSPS__LP_MODE_4_KSPS = 2
    HR_MODE_4_KSPS__LP_MODE_2_KSPS = 3
    HR_MODE_2_KSPS__LP_MODE_1_KSPS = 4
    HR_MODE_1_KSPS__LP_MODE_500_SPS = 5
    HR_MODE_500_SPS__LP_MODE_250_SPS = 6


@dataclass
class Ads129xConfigRegister1:
    """Describes config register 1 of ADS129x chip"""

    power_mode = Ads129xPowerMode.HIGH_RESOLUTION
    read_mode = Ads129xReadMode.DAISY_CHAIN
    clock_connection = Ads129xClockConnection.OSCILLATOR_CLOCK_OUTPUT_DISABLED
    output_data_rate = Ads129xOutputDataRate.HR_MODE_4_KSPS__LP_MODE_2_KSPS

    def set_data(self, data: bytes):
        """Convert data to information"""
        tmp = data[0]
        self.power_mode = Ads129xPowerMode((tmp >> 7) & 0x01)
        self.read_mode = Ads129xReadMode((tmp >> 6) & 0x01)
        self.clock_connection = Ads129xClockConnection((tmp >> 5) & 0x01)
        self.output_data_rate = Ads129xOutputDataRate((tmp >> 0) & 0x07)


    def get_data(self) -> bytes:
        """Convert information to bytes"""
        tmp = ((self.power_mode << 7) | (self.read_mode << 6) |
               (self.clock_connection << 5) | (self.output_data_rate << 0))
        return [tmp]



"""Provides a class for a usb connection"""

# import usb.core
# import usb.util
from .connection import Connection

class UsbConnection(Connection):
    """USB connection class,
    IMPORTANT: work in progress (there are driver issues under windows)"""


    # @staticmethod
    # def list_devices() -> list[usb.core.Device]:
    #     """Returns all USB devices"""
    #     return list(usb.core.find(find_all=True))


    # @staticmethod
    # def list_science_mode_devices() -> list[usb.core.Device]:
    #     """Returns all potential science mode USB devices"""
    #     devices = UsbConnection.list_devices()
    #     # science mode devices (P24/I24) have an STM32 mcu and these are
    #     # default values for USB CDC devices
    #     filtered_devices = list(filter(lambda x: x.idVendor == 0x0483 and x.idProduct == 0x5740 and
    #                                    x.bDeviceClass == 0x02, devices))
    #     return filtered_devices


    # def __init__(self, device: usb.core.Device):
    #     self._device = device
    #     self._out_endpoint = None
    #     self._in_endpoint = None
    #     self._is_open = False


    # def open(self):
    #     # P24/I24 have only one configuration
    #     self._device.set_configuration()
    #     # get an endpoint instance
    #     cfg = self._device.get_active_configuration()
    #     intf = cfg[(0,0)]

    #     # match the first OUT endpoint
    #     self._out_endpoint = usb.util.find_descriptor(
    #         intf,
    #         custom_match = lambda e: usb.util.endpoint_direction(e.bEndpointAddress) == usb.util.ENDPOINT_OUT)
    #     # match the first IN endpoint
    #     self._in_endpoint = usb.util.find_descriptor(
    #         intf,
    #         custom_match = lambda e: usb.util.endpoint_direction(e.bEndpointAddress) == usb.util.ENDPOINT_IN)

    #     self._is_open = True


    # def close(self):
    #     self._is_open = False


    # def is_open(self) -> bool:
    #     return self._is_open


    # def write(self, data: bytes):
    #     self._out_endpoint.write(data)


    # def read(self) -> bytes:
    #     return self._in_endpoint.read()


    # def clear_buffer(self):
    #     pass
